import os
from pathlib import Path


def project_root():
    cwd = Path(os.getcwd())
    while cwd != cwd.parent:
        if (cwd / "pyproject.toml").exists():
            return cwd
        cwd = cwd.parent
    return None



import os.path
from datetime import datetime
from os import PathLike

import pytest
from atomkraft.utils.project import project_root

from ..reactor.reactor import get_reactor

TRACE_TEST_STUB = """
from modelator.pytest.decorators import itf

pytest_plugins = "{0}"

@itf("{1}", keypath="{2}")
def test_trace():
    print("Successfully executed trace {1}")
"""


def test_trace(trace: PathLike, reactor: PathLike, keypath: str):
    """
    Test blockchain by running one trace
    """

    if reactor is None:
        reactor = get_reactor()

    root = project_root()
    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    tests = os.path.join(root, "tests")
    if not os.path.isdir(tests):
        raise RuntimeError(
            "Atomkraft tests directory not found: have you executed `atomkraft init`?"
        )

    timestamp = datetime.now().isoformat(timespec="milliseconds")
    test_name = f"test_{str(trace)}_{timestamp}"
    test_name = (
        test_name.replace("/", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
    )
    test_path = os.path.join(tests, f"{test_name}.py")
    with open(test_path, "w") as test:
        print(f"Writing {test_name} ...")
        test.write(
            TRACE_TEST_STUB.format(
                str(reactor).replace("/", ".").removesuffix(".py"), trace, keypath
            )
        )
    print(f"Executing {test_name} ...")
    pytest.main(["-s", test_path])



import ast
import os
import os.path
from os import PathLike
from typing import List, Optional

import tomlkit
from atomkraft.utils.project import project_root
from caseconverter import snakecase

from . import constants, utils
from .step_functions_visitor import StepFunctionsVisitor


def check_reactor(trace: PathLike, reactor=None) -> bool:
    """
    returns true if each action appearing in `trace` is matched with a function in `reactor`
    """
    if reactor is None:
        reactor = get_reactor()

    with open(reactor) as f_reactor:
        root_node = ast.parse(f_reactor.read())
    v = StepFunctionsVisitor()
    v.visit(root_node)

    keypath = utils.get_keypath(root_node)
    try:
        all_trace_actions = utils.get_all_trace_actions(trace, keypath)
    except Exception:
        raise ValueError(
            f"Keypath {keypath}, which is set in {reactor} is not present in the trace {trace}"
        )

    if constants.ALL_ENCOMPASSING_STEP in v.step_functions:
        return True
    else:
        return all_trace_actions.issubset(v.step_functions)


def get_reactor() -> PathLike:
    """
    returns the path to the current reactor from the internal config
    """

    root = project_root()

    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    internal_config_file_path = os.path.join(
        root,
        constants.ATOMKRAFT_INTERNAL_FOLDER,
        constants.ATOMKRAFT_INTERNAL_CONFIG,
    )

    if not os.path.isfile(internal_config_file_path):
        raise RuntimeError(
            "Atomkraft configuration not found: have you executed `atomkraft init`?"
        )

    with open(internal_config_file_path) as config_f:
        config_data = tomlkit.load(config_f)
        if constants.REACTOR_CONFIG_KEY not in config_data:
            raise RuntimeError(
                "Could not find default reactor; have you ran `atomkraft reactor`?"
            )
        return config_data[constants.REACTOR_CONFIG_KEY]


def generate_reactor(
    actions_list: List[str],
    variables_list: List[str],
    keypath: str = "action",
    stub_file_path: Optional[PathLike] = None,
) -> PathLike:

    imports_stub = _imports_stub()

    state_stub = _state_stub()
    actions_stub = "\n".join(
        [
            _action_stub(action_name=act, variables=variables_list)
            for act in actions_list
        ]
    )
    keypath_stub = _keypath_stub(keypath)
    with open(stub_file_path, "w") as f:
        f.write(imports_stub)
        f.write(keypath_stub)
        f.write(state_stub)
        f.write(actions_stub)

    if "PYTEST_CURRENT_TEST" in os.environ:
        root = "tests/project"
    else:
        root = project_root()

    internal_config_file_path = os.path.join(
        root,
        constants.ATOMKRAFT_INTERNAL_FOLDER,
        constants.ATOMKRAFT_INTERNAL_CONFIG,
    )

    atomkraft_internal_config = tomlkit.load(open(internal_config_file_path))
    atomkraft_internal_config[constants.REACTOR_CONFIG_KEY] = stub_file_path

    with open(internal_config_file_path, "w") as internal_config_f:
        tomlkit.dump(atomkraft_internal_config, internal_config_f)

    return stub_file_path


def _keypath_stub(keypath):
    stub = f"""

{constants.KEYPATH_VAR} = {repr(keypath)}
"""
    return stub


def _action_stub(action_name: str, variables: List[str]):
    stub = f"""


@step({repr(action_name)})
def {snakecase(action_name)}(testnet, state, {", ".join(variables)}):
    print("Step: {action_name}")
"""
    return stub


def _state_stub():
    stub = """


@pytest.fixture
def state():
    return {}
"""
    return stub


def _imports_stub():
    stub = """
import pytest
from modelator.pytest.decorators import step
"""
    return stub



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['atomkraft',
 'atomkraft.chain',
 'atomkraft.cli',
 'atomkraft.reactor',
 'atomkraft.test',
 'atomkraft.utils']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.27,<4.0.0',
 'bip-utils>=2.3.0,<3.0.0',
 'case-converter>=1.1.0,<2.0.0',
 'copier>=6.1.0,<7.0.0',
 'hdwallet>=2.1.1,<3.0.0',
 'jsonrpcclient>=4.0.2,<5.0.0',
 'modelator>=0.5.1,<0.6.0',
 'numpy>=1.22.4,<2.0.0',
 'pytest>=7.1.2,<8.0.0',
 'tabulate>=0.8.9,<0.9.0',
 'tomlkit>=0.11.1,<0.12.0',
 'typer[all]>=0.6.1,<0.7.0',
 'websockets>=10.3,<11.0']

entry_points = \
{'console_scripts': ['atomkraft = atomkraft.cli:app'],
 'pytest11': ['atomkraft-chain = atomkraft.chain.pytest']}

setup_kwargs = {
    'name': 'atomkraft',
    'version': '0.0.4',
    'description': 'Testing for Cosmos Blockchains',
    'long_description': '# atomkraft\n\nTesting for Cosmos Blockchains\n\n### Using `pip` (inside a system or virtual env)\n\n```\npip install atomkraft\natomkraft --help\n# or\npython -m atomkraft --help\n```\n\n### Using `poetry` (inside a project)\n\n```\npoetry add atomkraft\npoerty run atomkraft --help\n# or\npoetry run python -m atomkraft --help\n```\n\n### Code Quality\n\n```\npip install black pylama[all]\nblack . --check\npylama -l pyflakes,pycodestyle,isort\n```\n',
    'author': 'Andrey Kuprianov',
    'author_email': 'andrey@informal.systems',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/informalsystems/atomkraft',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
